/* $Id: diva_didd.c,v 1.1.2.5 2001/02/16 08:40:36 armin Exp $
 *
 * DIDD Interface module for Eicon active cards.
 * 
 * Functions are in dadapter.c 
 * 
 * Copyright 2000-2007 by Armin Schindler (armin@melware.de) 
 * Copyright 2000-2007 Cytronics & Melware (info@melware.de)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/poll.h>
#include <linux/pci.h>

#include "platform.h"
#include "di_defs.h"
#include "dadapter.h"
#include "divasync.h"
#include "did_vers.h"
#include "cfg_types.h"
#include "cfg_ifc.h"
#include "cfg_notify.h"


static char *main_revision = "$Revision: 1.1.2.5 $";

static char *DRIVERNAME = "Eicon DIVA - DIDD table (http://www.melware.net)";
static char *DRIVERLNAME = "divadidd";
static char *DRIVERRELEASE = "3.0.10-107.884-1";

static char *dir_in_proc_net = "isdn";
static char *main_proc_dir = "eicon";

MODULE_DESCRIPTION(             "DIDD table driver for diva drivers");
MODULE_AUTHOR(                  "Cytronics & Melware, Eicon Networks");
MODULE_SUPPORTED_DEVICE(        "Eicon diva drivers");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#define DBG_MINIMUM  (DL_LOG + DL_FTL + DL_ERR)
#define DBG_DEFAULT  (DBG_MINIMUM + DL_XLOG + DL_REG)


extern void DIVA_DIDD_Read (void *, int);
static dword notify_handle;
static DESCRIPTOR _DAdapter;

static struct proc_dir_entry *proc_net_isdn;
static struct proc_dir_entry *proc_didd;
struct proc_dir_entry *proc_net_isdn_eicon = NULL;

static loff_t  cfglib_lseek (struct file *file, loff_t offset, int orig);
static ssize_t cfglib_read (struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t cfglib_write (struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int cfglib_poll (struct file *file, poll_table *wait);
static ssize_t cfglib_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static ssize_t cfglib_open (struct inode *inode, struct file *file);
static ssize_t cfglib_release (struct inode *inode, struct file *file);
static struct semaphore didd_ifc_lock;

typedef struct _diva_cfglib_application {
  atomic_t              read_ready;
  void*                 notify_handle;
  diva_section_target_t target;
  wait_queue_head_t     read_wait;
  char*                 didd_info_buffer;
	int 									didd_info_offset;
	int										didd_info_length;
} diva_cfglib_application_t;

EXPORT_SYMBOL_NOVERS(DIVA_DIDD_Read);
EXPORT_SYMBOL_NOVERS(proc_net_isdn_eicon);

static char *
getrev(const char *revision)
{
	char *rev;
	char *p;
	if ((p = strchr(revision, ':'))) {
		rev = p + 2;
		p = strchr(rev, '$');
		*--p = 0;
	} else rev = "1.0";
	return rev;
}

/* -------------------------------------------------------
    Spin Lock framework
   ------------------------------------------------------- */
void diva_os_enter_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  spin_lock(a);
}

void diva_os_leave_spin_lock (diva_os_spin_lock_t* a,
                              diva_os_spin_lock_magic_t* old_irql,
                              void* dbg) {
  spin_unlock(a);
}

/*
** proc idi entries
*/
struct file_operations cfglib_fops = {
  llseek:      cfglib_lseek,
  read:        cfglib_read,
  write:       cfglib_write,
  poll:        cfglib_poll,
  ioctl:       cfglib_ioctl,
  open:        cfglib_open,
  release:     cfglib_release,
};


static int __init create_proc(void)
{
  struct proc_dir_entry *pe;

  for (pe = proc_net->subdir; pe; pe = pe->next) {
    if (!memcmp(dir_in_proc_net, pe->name, pe->namelen)) {
        proc_net_isdn = pe;
        break;
    }
  }
  if (!proc_net_isdn) {
    proc_net_isdn = create_proc_entry(dir_in_proc_net, S_IFDIR, proc_net);
  }
  proc_net_isdn_eicon = create_proc_entry(main_proc_dir, S_IFDIR, proc_net_isdn);

  if(proc_net_isdn_eicon) {
    if((proc_didd = create_proc_entry(DRIVERLNAME, S_IFREG | S_IRUGO, proc_net_isdn_eicon))) {
      proc_didd->proc_fops = &cfglib_fops;
      proc_didd->owner = THIS_MODULE;
    }
    return(1);
  }

  return(0);
}

static void __exit remove_proc(void)
{
  remove_proc_entry(DRIVERLNAME, proc_net_isdn_eicon);
  remove_proc_entry(main_proc_dir, proc_net_isdn);

  if ((proc_net_isdn) && (!proc_net_isdn->subdir)){
    remove_proc_entry(dir_in_proc_net, proc_net);
  }
}

static void *
didd_callback(void *context, DESCRIPTOR* adapter, int removal)
{
  if (adapter->type == IDI_DADAPTER)
  {
    printk(KERN_ERR "%s: Change in DAdapter ? Oops ?.\n", DRIVERLNAME);
    DBG_ERR(("Notification about IDI_DADAPTER change ! Oops."));
    return(NULL);
  }
  else if (adapter->type == IDI_DIMAINT)
  {
    if (removal) {
      DbgDeregister();
    } else {
      DbgRegister("DIDD", DRIVERRELEASE, DBG_DEFAULT);
    }
  } else if (adapter->channels != 0       &&
						 adapter->type < IDI_VADAPTER &&
						 adapter->features != 0       &&
						 adapter->request) {
		if (removal == 0) {
			diva_cfg_lib_notify_adapter_insertion(adapter->request);
		}
	}
  return(NULL);
}

/* -------------------------------------------------------------------------
  Portable debug Library
  ------------------------------------------------------------------------- */
#include "debuglib.c"
/* ------------------------------------------------------------------------- */

static int __init
connect_didd(void)
{
  int x = 0;
  int dadapter = 0;
  IDI_SYNC_REQ req;
  DESCRIPTOR DIDD_Table[MAX_DESCRIPTORS];

  DIVA_DIDD_Read(DIDD_Table, sizeof(DIDD_Table));

  for (x = 0; x < MAX_DESCRIPTORS; x++)
  {
    if (DIDD_Table[x].type == IDI_DADAPTER)
    {  /* DADAPTER found */
      dadapter = 1;
      memcpy(&_DAdapter, &DIDD_Table[x], sizeof(_DAdapter));
      req.didd_notify.e.Req = 0;
      req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REGISTER_ADAPTER_NOTIFY;
      req.didd_notify.info.callback = didd_callback;
      req.didd_notify.info.context = 0;
      _DAdapter.request((ENTITY *)&req);
      if (req.didd_notify.e.Rc != 0xff)
        return(0);
      notify_handle = req.didd_notify.info.handle;
    }
    else if (DIDD_Table[x].type == IDI_DIMAINT)
    {  /* MAINT found */
      DbgRegister("DIDD", DRIVERRELEASE, DBG_DEFAULT);
    }
  }
  return(dadapter);
}

static void __exit
disconnect_didd(void)
{
  IDI_SYNC_REQ req;

  req.didd_notify.e.Req = 0;
  req.didd_notify.e.Rc = IDI_SYNC_REQ_DIDD_REMOVE_ADAPTER_NOTIFY;
  req.didd_notify.info.handle = notify_handle;
  _DAdapter.request((ENTITY *)&req);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#define dma_addr_t unsigned long

#include <asm/page.h>
/* Pure 2^n version of get_order */
static int diva_get_order(unsigned long size)
{
	int order;

	size = (size-1) >> (PAGE_SHIFT-1);
	order = -1;
	do {
		size >>= 1;
		order++;
	} while (size);
	return order;
}

static void *diva_pci_alloc_consistent(struct pci_dev *hwdev,
                                       size_t size,
                                       dma_addr_t * dma_handle,
                                       void** addr_handle) {
  void *ret  = (void *)__get_free_pages(GFP_ATOMIC, diva_get_order(size));

  *addr_handle = ret;

	if (ret != NULL) {
#ifdef	__mips__
    dma_cache_wback_inv (ret, diva_get_order(size)*PAGE_SIZE);
    ret = (void*)vtonocache(ret);
#endif
		memset(ret, 0, size);
		*dma_handle = virt_to_bus(ret);
	}

	return ret;
}

static void pci_free_consistent(struct pci_dev *hwdev, size_t size,
			 void *vaddr, dma_addr_t dma_handle)
{
	unsigned long addr = (unsigned long) vaddr;

	free_pages(addr, diva_get_order(size));
}

#else

static void *diva_pci_alloc_consistent (struct pci_dev* hwdev,
                                        size_t size,
                                        dma_addr_t * dma_handle,
                                        void** addr_handle) {
  void* addr = pci_alloc_consistent (0, size, dma_handle);

  *addr_handle = addr;

  return (addr);
}

#endif

static void*      clock_data_addr;
static void*      clock_data_addr_handle;
static dma_addr_t clock_data_bus_addr;

int diva_get_clock_data (dword* bus_addr_lo,
                         dword* bus_addr_hi,
                         dword* length,
                         void** addr) {
	if (clock_data_addr != 0 &&
			(sizeof(clock_data_bus_addr) == sizeof(dword) || ((dword)(clock_data_bus_addr >> 32)) == 0)) {
		*bus_addr_lo = (dword)clock_data_bus_addr;
		*addr        = clock_data_addr;
		*length      = PAGE_SIZE;

		return (0);
	}

  return (-1);
}

static int __init
divadidd_init(void)
{
  char tmprev[32];
  int ret = 0;

  MOD_INC_USE_COUNT;

  printk(KERN_INFO "%s\n", DRIVERNAME);
  printk(KERN_INFO "%s: Rel:%s  Rev:", DRIVERLNAME, DRIVERRELEASE);
  strcpy(tmprev, main_revision);
  printk("%s  Build:%s(%s)\n", getrev(tmprev),
        diva_didd_common_code_build, DIVA_BUILD);

  init_MUTEX_LOCKED(&didd_ifc_lock);

  if (!create_proc()) {
    printk(KERN_ERR "%s: could not create proc entry\n", DRIVERLNAME);
    ret = -EIO;
    goto out;
  }

  diva_didd_load_time_init();
  if (diva_cfg_lib_init ()) {
    printk(KERN_ERR "%s: failed to init CFGLib.\n", DRIVERLNAME);
    diva_didd_load_time_finit();
    remove_proc();
    ret = -EIO;
    goto out;
  }

  if(!connect_didd()) {
    printk(KERN_ERR "%s: failed to connect to DIDD.\n", DRIVERLNAME);
    diva_cfg_lib_finit ();
    diva_didd_load_time_finit();
    remove_proc();
    ret = -EIO;
    goto out;
  }

  up(&didd_ifc_lock);

	if ((clock_data_addr =
			diva_pci_alloc_consistent (0, PAGE_SIZE, &clock_data_bus_addr, &clock_data_addr_handle)) != 0)
		memset (clock_data_addr, 0x00, PAGE_SIZE);

out:
  MOD_DEC_USE_COUNT;
  return (ret);
}

static loff_t  cfglib_lseek (struct file *file, loff_t offset, int orig) {
  return (-ESPIPE);
}

/*
  Read depends on the type of application:
  - configuration response
  - configuration update notification
  */
static ssize_t cfglib_read (struct file *file, char *buf, size_t count, loff_t *offset) {
  diva_cfglib_application_t* appl;
  int retval = 0;

  if (offset != &file->f_pos)
    return (-ESPIPE);

  down(&didd_ifc_lock);

  appl = (diva_cfglib_application_t*)file->private_data;

  if (appl->notify_handle) {
    if (atomic_read (&appl->read_ready) != 0) {
      if (appl->target == TargetCFGLibDrv) {
        int length = diva_cfg_get_read_buffer_length ();

        if (length && (count >= (size_t)length)) {
          char* data = 0;

          atomic_set (&appl->read_ready, 0);

          length = diva_cfg_lib_read (&data, length);

          if (data) {
            if ((retval = copy_to_user (buf, data, length)) == 0) {
              retval = length;
            }
            diva_os_free (0, data);
          } else {
            retval = -EIO;
          }
        } else {
          retval = -EMSGSIZE;
        }
      } else {
				/*
					Notification read by application
					*/
				byte* data;
				int length = diva_cfg_lib_get_owner_data (&data, appl->target);
				if (length > 0) {
					if (count >= (size_t)length) {
            if ((retval = copy_to_user (buf, data, length)) == 0) {
              retval = length;
							atomic_set (&appl->read_ready, 0);
            }
					} else {
						retval = -EMSGSIZE;
					}
				} else {
					retval = -EIO;
				}
      }
    } else {
      retval = -EIO;
    }
  } else {
    /*
      Request for information about DIDD version
      */
    if (appl->didd_info_buffer == 0) {
      int length = 0, max_length = 1024;
      char* dst = vmalloc (max_length);

      if ((appl->didd_info_buffer = dst)) {
        char tmprev[32];

        diva_snprintf (tmprev, sizeof(tmprev), "%s", main_revision);
        length += diva_snprintf (&dst[length], max_length - length, "%s\n", DRIVERNAME);
        length += diva_snprintf (&dst[length], max_length - length, "name     : %s\n", DRIVERLNAME);
        length += diva_snprintf (&dst[length], max_length - length, "build    : %s(%s)\n",
                                 diva_didd_common_code_build, DIVA_BUILD);
        length += diva_snprintf (&dst[length], max_length - length, "revision : %s\n", getrev(tmprev));
        length += diva_snprintf (&dst[length], max_length - length, "cfglib   : %d\n", VieLlast);
        length += diva_snprintf (&dst[length], max_length - length, "cfgstate : %d\n",
																 (diva_cfg_lib_get_storage () != 0) ? 1 : 0);

        appl->didd_info_offset = 0;
        appl->didd_info_length = length;
      }
    }

    if (appl->didd_info_buffer) {
      int to_copy = MIN(count, appl->didd_info_length);

      if (to_copy &&
          (retval = copy_to_user (buf, &appl->didd_info_buffer[appl->didd_info_offset], to_copy)) == 0) {
        appl->didd_info_offset += to_copy;
        appl->didd_info_length -= to_copy;
        retval = to_copy;
      }
    } else {
      retval = -ENOMEM;
    }
  }

  up(&didd_ifc_lock);

  return (retval);
}

static int cfg_lib_notify (void* context, const byte* message, const byte* instance) {
  diva_cfglib_application_t* appl = (diva_cfglib_application_t*)context;

	if (message == 0) {
  	atomic_set (&appl->read_ready, 1);
  	wake_up_interruptible (&appl->read_wait);

  	return ((appl->target == TargetCFGLibDrv) ? 0 : 1);
	} else {
		return (-1);
	}
}

/*
  Write operation. Three write formats are supported:
  - configuration update format, context not necessary
  - user mode application register for configuration change
    notifications
  - user mode application configuration update response
  */
static ssize_t cfglib_write (struct file *file, const char *buf, size_t count, loff_t *offset) {
  diva_cfglib_application_t* appl;
  int retval = 0;
  byte* data;

  if (offset != &file->f_pos)
    return (-ESPIPE);

  if (count < 4)
    return (-EINVAL);

  if (!(data = (byte*)vmalloc ((unsigned int)count+1)))
    return -ENOMEM;

  if ((retval = copy_from_user(data, buf, count))) {
    vfree (data);
    return (retval);
  }

  data[count] = 0;

  if (!(appl = (diva_cfglib_application_t*)file->private_data)) {
    vfree (data);
    return (-ENOMEM);
  }

  down(&didd_ifc_lock);

  if (appl->notify_handle == 0) {
    /*
      No application is bound to this file descriptor
      */

    if (data[0] == '<') {
      appl->target = TargetCFGLibDrv;
    } else if (count == 4) {
      appl->target = (diva_section_target_t)READ_DWORD(data);
    } else {
      retval = -EINVAL;
    }
    if (retval == 0) {
      if ((appl->notify_handle = diva_cfg_lib_cfg_register_notify_callback (cfg_lib_notify,
                                                                            appl, appl->target))) {
        if (appl->target == TargetCFGLibDrv) {
          if (diva_cfg_lib_write (data)) {
            retval = -EINVAL;
          }
        }
      } else {
        retval = -EINVAL;
      }
    }
  } else {
   /*
     Process application specific request from existing application
     */
    if (appl->target == TargetCFGLibDrv) {
      if (diva_cfg_lib_write (data)) {
        retval = -EINVAL;
      }
    } else if (count == 4) {
      /*
        Application configuration change response
        */
			byte* notify_data;

			diva_cfg_lib_ack_owner_data (appl->target, READ_DWORD(data));
			if (diva_cfg_lib_get_owner_data (&notify_data, appl->target) > 0) {
				atomic_set (&appl->read_ready, 1);
			}
    } else {
      retval = -EINVAL;
    }
  }

  up(&didd_ifc_lock);

  vfree (data);

  return ((retval != 0) ? (ssize_t)retval : (ssize_t)count);
}

static unsigned int cfglib_poll (struct file *file, poll_table *wait) {
  diva_cfglib_application_t* appl;
  unsigned int retval = 0;

  appl = (diva_cfglib_application_t*)file->private_data;

  if (appl->notify_handle) {
  	poll_wait (file, &appl->read_wait, wait);
    if (atomic_read (&appl->read_ready) != 0) {
      retval = (POLLIN | POLLRDNORM);
    }
  } else {
    retval = POLLERR;
  }

  return (retval);
}

static ssize_t cfglib_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) {
  return (-ENODEV);
}

static ssize_t cfglib_open (struct inode *inode, struct file *file) {
  diva_cfglib_application_t* appl = (diva_cfglib_application_t*)vmalloc(sizeof(*appl));

  if (appl) {
    memset (appl, 0x00, sizeof(*appl));

    appl->target = TargetLast;
    atomic_set (&appl->read_ready, 0);
		init_waitqueue_head(&appl->read_wait);

    file->private_data = appl;
  } else {
    file->private_data = 0;
    return (-ENOMEM);
	}

  return (0);
}

static ssize_t cfglib_release (struct inode *inode, struct file *file) {
  diva_cfglib_application_t* appl;

  down(&didd_ifc_lock);

  if ((appl = (diva_cfglib_application_t*)file->private_data)) {
    file->private_data = 0;

    if (appl->notify_handle != 0) {
      diva_cfg_lib_cfg_remove_notify_callback (appl->notify_handle);
    }
    if (appl->didd_info_buffer) {
      vfree (appl->didd_info_buffer);
    }

    vfree (appl);
  }

  up(&didd_ifc_lock);


  return (0);
}

void diva_os_read_descriptor_array (void* t, int length) {
	DIVA_DIDD_Read (t, length);
}

void
diva_os_sleep (dword mSec)
{
  unsigned long timeout = HZ * mSec / 1000 + 1;

  set_current_state(TASK_UNINTERRUPTIBLE);
  schedule_timeout(timeout);
}

/*********************************************************
 ** malloc / free
 *********************************************************/

void* diva_os_malloc (unsigned long flags, unsigned long size)
{
  void* ret = NULL;
 
  if (size) {
      ret = (void*)vmalloc ((unsigned int)size);
  }
  return (ret);
}

void
diva_os_free (unsigned long unused, void* ptr)
{
  if (ptr) {
    vfree (ptr);
  }
}

void __exit
divadidd_exit(void)
{
  diva_cfg_lib_finit ();
  DbgDeregister();
  disconnect_didd();
  diva_didd_load_time_finit();
  remove_proc();

	if (clock_data_addr_handle != 0) {
		pci_free_consistent (0, PAGE_SIZE, clock_data_addr_handle, clock_data_bus_addr);
	}

  printk(KERN_INFO "%s: module unloaded.\n", DRIVERLNAME);
}

module_init(divadidd_init);
module_exit(divadidd_exit);
