#include "platform.h"
#include <stdarg.h>
#include "debuglib.h"
#include "cardtype.h"
#include "dlist.h"
#include "pc.h"
#include "pr_pc.h"
#include "di_defs.h"
#include "dsp_defs.h"
#include "di.h" 
#include "io.h"

#include "xdi_msg.h"
#include "xdi_adapter.h"
#include "os_analog.h"
#include "diva_pci.h"
#include "mi_pc.h"
#include "dsrv_analog.h"
#include "dsrv4bri.h"

/*
**  IMPORTS
*/
extern void* diva_xdiLoadFileFile;
extern dword diva_xdiLoadFileLength;
extern void diva_xdi_display_adapter_features (int card);
extern int qBri_FPGA_download (PISDN_ADAPTER IoAdapter);
extern void start_analog_hardware (PISDN_ADAPTER IoAdapter);
extern int diva_card_read_xlog (diva_os_xdi_adapter_t* a);
extern void prepare_analog_functions (PISDN_ADAPTER IoAdapter);

/*
**  LOCALS
*/
static unsigned long _analog_bar_length[4] = {
  512,
  0x100, /* I/O */
  MQ2_MEMORY_SIZE,
  256*1024
};


static int diva_analog_cleanup_adapter (diva_os_xdi_adapter_t* a);
static int diva_analog_cmd_card_proc (struct _diva_os_xdi_adapter* a,
                                    diva_xdi_um_cfg_cmd_t* cmd,
                                    int length);
static int diva_analog_write_fpga_image (diva_os_xdi_adapter_t* a,
                                       byte* data,
                                       dword length);
static int diva_analog_reset_adapter (struct _diva_os_xdi_adapter* a);
static int diva_analog_write_sdram_block (PISDN_ADAPTER IoAdapter,
                                        dword address,
                                        const byte* data,
                                        dword length,
                                        dword limit);
static int diva_analog_start_adapter (PISDN_ADAPTER IoAdapter,
                                    dword start_address,
                                    dword features);
static int check_analog_interrupt (PISDN_ADAPTER IoAdapter);
static int diva_analog_stop_adapter (diva_os_xdi_adapter_t* a);

/*
**  BAR0 - MEM - 512             - CONFIG MEM
**  BAR1 - I/O - 0x100           - UNUSED
**  BAR2 - MEM - MQ2_MEMORY_SIZE - SDRAM
**  BAR3 - MEM - 256KBytes       - CNTRL
**
**  Called by master adapter, that will initialize and add slave adapters
*/
int
diva_analog_init_card (diva_os_xdi_adapter_t* a)
{
  int bar;
  unsigned long bar_length[sizeof(_analog_bar_length)/sizeof(_analog_bar_length[0])];
	memcpy (bar_length, _analog_bar_length, sizeof(bar_length));

  /*
    Fix up card ordinal to get righ adapter name and features
    */
  {
    byte Bus, Slot;
    void* hdev;
    word id = 0;

    Bus = a->resources.pci.bus;
    Slot = a->resources.pci.func;
    hdev = a->resources.pci.hdev;

    PCIread (Bus, Slot, 0x2e, &id, sizeof(id), hdev) ;
    DBG_LOG(("A(%d) read Subsystem ID %04x", a->xdi_adapter.ANum, id))

		if (a->CardOrdinal == CARDTYPE_DIVASRV_ANALOG_8PORT) {
			switch (id) {
				case 0x2800:
					a->CardOrdinal = CARDTYPE_DIVASRV_V_ANALOG_8PORT;
					break;
			}
		} else if (a->CardOrdinal == CARDTYPE_DIVASRV_ANALOG_4PORT) {
			switch (id) {
				case 0x2400:
					a->CardOrdinal = CARDTYPE_DIVASRV_V_ANALOG_4PORT;
					break;
			}
		} else if (a->CardOrdinal == CARDTYPE_DIVASRV_ANALOG_2PORT) {
			switch (id) {
				case 0x2200:
					a->CardOrdinal = CARDTYPE_DIVASRV_V_ANALOG_2PORT;
					break;
			}
		}
  }

	a->xdi_adapter.BusNumber  = a->resources.pci.bus;
	a->xdi_adapter.slotNumber = a->resources.pci.func;
	a->xdi_adapter.hdev       = a->resources.pci.hdev;
  /*
    Get Serial Number
    The serial number of PLX is accessible in accordance with PCI spec
    via command register located in configuration space, also we do not
    have to map any BAR before we can access it
  */
	if (!diva_get_serial_number (&a->xdi_adapter)) {
    DBG_ERR(("ANALOG: can't ger Serial Number"))
    diva_analog_cleanup_adapter (a);
    return (-1);
  }

  /*
    Set properties
  */
  a->xdi_adapter.Properties = CardProperties[a->CardOrdinal];
  DBG_LOG(("Load %s, SN:%ld, bus:%02x, func:%02x",
            a->xdi_adapter.Properties.Name,
            a->xdi_adapter.serialNo,
            a->resources.pci.bus,
            a->resources.pci.func))

  /*
    First initialization step: get and check hardware resoures.
    Do not map resources and do not access card at this step
  */

  for (bar = 0; bar < 4; bar++) {
    a->resources.pci.bar[bar] = divasa_get_pci_bar (a->resources.pci.bus,
                                                    a->resources.pci.func,
                                                    bar,
                                                    a->resources.pci.hdev);
    if (!a->resources.pci.bar[bar] ||
        (a->resources.pci.bar[bar] == 0xFFFFFFF0)) {
      DBG_ERR(("A: invalid bar[%d]=%08x", bar, a->resources.pci.bar[bar]))
      return (-1);
    }
  }
  a->resources.pci.irq = (byte)divasa_get_pci_irq (a->resources.pci.bus,
                                                   a->resources.pci.func,
                                                   a->resources.pci.hdev);
  if (!a->resources.pci.irq) {
    DBG_ERR(("A: invalid irq"));
    return (-1);
  }

  a->xdi_adapter.sdram_bar = a->resources.pci.bar [2];

  /*
    Map all MEMORY BAR's
  */
  for (bar = 0; bar < 4; bar++) {
    if (bar != 1) { /* ignore I/O */
      a->resources.pci.addr[bar] = divasa_remap_pci_bar (a, a->resources.pci.bar[bar],
                                                         bar_length[bar]);
      if (!a->resources.pci.addr[bar]) {
        DBG_ERR(("ANALOG: can't map bar[%d]", bar))
        diva_analog_cleanup_adapter (a);
        return (-1);
      }
    }
  }

  /*
    Register I/O port
  */
  sprintf (&a->port_name[0], "DIVA ANALOG %ld", (long)a->xdi_adapter.serialNo);

  if (diva_os_register_io_port (a, 1, a->resources.pci.bar[1],
                                bar_length[1],
                                &a->port_name[0])) {
    DBG_ERR(("ANALOG: can't register bar[1]"))
    diva_analog_cleanup_adapter (a);
    return (-1);
  }

  a->resources.pci.addr[1] = (void*)(unsigned long)a->resources.pci.bar[1];

  /*
    Set cleanup pointer for base adapter only, so slave adapter
    will be unable to get cleanup
  */
  a->interface.cleanup_adapter_proc = diva_analog_cleanup_adapter;

  a->xdi_adapter.a.io = &a->xdi_adapter;
  a->xdi_adapter.DIRequest = request;
  a->interface.cmd_proc = diva_analog_cmd_card_proc;

  a->xdi_adapter.Channels = CardProperties[a->CardOrdinal].Channels;
  a->xdi_adapter.e_max = CardProperties[a->CardOrdinal].E_info;
  a->xdi_adapter.e_tbl = diva_os_malloc (0, a->xdi_adapter.e_max * sizeof(E_INFO)); 

  if (!a->xdi_adapter.e_tbl) {
    diva_analog_cleanup_adapter (a);
    return (-1);
  }
  memset (a->xdi_adapter.e_tbl, 0x00, a->xdi_adapter.e_max * sizeof(E_INFO));

  a->xdi_adapter.cardType = a->CardOrdinal;

  if (diva_os_initialize_spin_lock (&a->xdi_adapter.isr_spin_lock, "isr")) {
    diva_analog_cleanup_adapter (a);
    return (-1);
  }
  if (diva_os_initialize_spin_lock (&a->xdi_adapter.data_spin_lock, "data")) {
    diva_analog_cleanup_adapter (a);
    return (-1);
  }

  if (a->CardOrdinal == CARDTYPE_DIVASRV_ANALOG_2PORT ||
      a->CardOrdinal == CARDTYPE_DIVASRV_V_ANALOG_2PORT) {
    a->dsp_mask = 0x00000003;
    strcpy (a->xdi_adapter.req_soft_isr.dpc_thread_name, "kdivas2and");
  }  
  else if (a->CardOrdinal == CARDTYPE_DIVASRV_ANALOG_4PORT ||
      a->CardOrdinal == CARDTYPE_DIVASRV_V_ANALOG_4PORT) {
    a->dsp_mask = 0x0000000f;
    strcpy (a->xdi_adapter.req_soft_isr.dpc_thread_name, "kdivas4and");
  } else {
    strcpy (a->xdi_adapter.req_soft_isr.dpc_thread_name, "kdivas8and");
    a->dsp_mask = 0x000000ff;
  }

  if (diva_os_initialize_soft_isr (&a->xdi_adapter.req_soft_isr,
                                   DIDpcRoutine,
                                   &a->xdi_adapter)) {
    diva_analog_cleanup_adapter (a);
    return (-1);
  }

  /*
   Do not initialize second DPC - only one thread will be created
  */
  a->xdi_adapter.isr_soft_isr.object = a->xdi_adapter.req_soft_isr.object;
  prepare_analog_functions (&a->xdi_adapter);

  diva_init_dma_map (a->resources.pci.hdev, (struct _diva_dma_map_entry**)&a->xdi_adapter.dma_map, 32);

  /*
    Set up hardware related pointers
  */
  a->xdi_adapter.cfg = (void*)(unsigned long)a->resources.pci.bar[0]; /* BAR0 CONFIG */
  a->xdi_adapter.port = (void*)(unsigned long)a->resources.pci.bar[1]; /* BAR1        */
  a->xdi_adapter.Address = a->resources.pci.addr[2];        /* BAR2 SDRAM  */
  a->xdi_adapter.ctlReg = (void*)(unsigned long)a->resources.pci.bar[3]; /* BAR3 CNTRL  */

  a->xdi_adapter.reset = a->resources.pci.addr[0]; /* BAR0 CONFIG */
  a->xdi_adapter.ram = a->resources.pci.addr[2]; /* BAR2 SDRAM  */
  a->xdi_adapter.ram  += MP_SHARED_RAM_OFFSET;

  a->xdi_adapter.AdapterTestMemoryStart  = a->resources.pci.addr[2]; /* BAR2 SDRAM  */
  a->xdi_adapter.AdapterTestMemoryLength = bar_length[2] - 1;

  /*
    ctlReg contains the register address for the MIPS CPU reset control
  */
  a->xdi_adapter.ctlReg = a->resources.pci.addr[3];  /* BAR3 CNTRL  */
  /*
    prom contains the register address for FPGA and EEPROM programming
  */
  a->xdi_adapter.prom = &a->xdi_adapter.reset[0x6E];
  /*
    reset contains the base address for the PLX 9054 register set
  */
  a->xdi_adapter.reset[PLX9054_INTCSR] = 0x00 ;	/* disable PCI interrupts */


  /*
    Set IRQ handler
  */
  a->xdi_adapter.irq_info.irq_nr = a->resources.pci.irq;
  sprintf (a->xdi_adapter.irq_info.irq_name, "DIVA ANALOG %ld", (long)a->xdi_adapter.serialNo);

  if (diva_os_register_irq (a, a->xdi_adapter.irq_info.irq_nr,
                            a->xdi_adapter.irq_info.irq_name, a->xdi_adapter.msi)) {
    diva_analog_cleanup_adapter (a);
    return (-1);
  }

  a->xdi_adapter.irq_info.registered = 1;

  diva_log_info("%s IRQ:%d SerNo:%d", a->xdi_adapter.Properties.Name,
                                      a->resources.pci.irq,
                                      a->xdi_adapter.serialNo);
  return (0);
}

/*
**  Cleanup function
*/
static int
diva_analog_cleanup_adapter (diva_os_xdi_adapter_t* a)
{
  int bar;

  /*
    Stop adapter if running
  */
  if (a->xdi_adapter.Initialized) {
    diva_analog_stop_adapter (a);
  }

  /*
    Remove IRQ handler
  */
  if (a->xdi_adapter.irq_info.registered) {
    diva_os_remove_irq (a, a->xdi_adapter.irq_info.irq_nr);
  }
  a->xdi_adapter.irq_info.registered = 0;

  diva_os_cancel_soft_isr (&a->xdi_adapter.req_soft_isr);
  diva_os_cancel_soft_isr (&a->xdi_adapter.isr_soft_isr);

  diva_os_remove_soft_isr (&a->xdi_adapter.req_soft_isr);
  a->xdi_adapter.isr_soft_isr.object = 0;

  diva_os_destroy_spin_lock (&a->xdi_adapter.isr_spin_lock, "unload");
  diva_os_destroy_spin_lock (&a->xdi_adapter.data_spin_lock,"unload");

  if (a->xdi_adapter.e_tbl) {
    diva_os_free (0, a->xdi_adapter.e_tbl);
  }
  a->xdi_adapter.e_tbl   = 0;
  a->xdi_adapter.e_max   = 0;
  a->xdi_adapter.e_count = 0;

  /*
    Unmap all BARS
  */
  for (bar = 0; bar < 4; bar++) {
    if (bar != 1) {
      if (a->resources.pci.bar[bar] && a->resources.pci.addr[bar]) {
        divasa_unmap_pci_bar (a->resources.pci.addr[bar]);
        a->resources.pci.bar[bar]  = 0;
        a->resources.pci.addr[bar] = 0;
      }
    }
  }

  /*
    Unregister I/O
  */
  if (a->resources.pci.bar[1] && a->resources.pci.addr[1]) {
    diva_os_register_io_port (a, 0, a->resources.pci.bar[1], 
      _analog_bar_length[1],
      &a->port_name[0]);
    a->resources.pci.bar[1]  = 0;
    a->resources.pci.addr[1] = 0;
  }

  diva_free_dma_map (a->resources.pci.hdev, (struct _diva_dma_map_entry*)a->xdi_adapter.dma_map);
  a->xdi_adapter.dma_map = 0;

  DBG_LOG(("A(%d) %s adapter cleanup complete", a->xdi_adapter.ANum, a->xdi_adapter.Properties.Name))

  return (0);
}

static int
diva_analog_cmd_card_proc (struct _diva_os_xdi_adapter* a,
                         diva_xdi_um_cfg_cmd_t* cmd,
                         int length)
{
  int ret = -1;

  if (cmd->adapter != a->controller) {
    DBG_ERR(("analog_cmd, invalid controller=%d != %d", cmd->adapter, a->controller))
    return (-1);
  }

  switch (cmd->command) {
    case DIVA_XDI_UM_CMD_GET_CARD_ORDINAL:
      a->xdi_mbox.data_length = sizeof(dword);
      a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length);
      if (a->xdi_mbox.data) {
        *(dword*)a->xdi_mbox.data = (dword)a->CardOrdinal;
        a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
        ret = 0;
      }
      break;

    case DIVA_XDI_UM_CMD_GET_SERIAL_NR:
      a->xdi_mbox.data_length = sizeof(dword);
      a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length);
      if (a->xdi_mbox.data) {
        *(dword*)a->xdi_mbox.data = (dword)a->xdi_adapter.serialNo;
        a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
        ret = 0;
      }
      break;

    case DIVA_XDI_UM_CMD_GET_PCI_HW_CONFIG:
      if (!a->xdi_adapter.ControllerNumber) {
        /*
          Only master adapter can access hardware config
        */
        a->xdi_mbox.data_length = sizeof(dword)*9;
        a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length);
        if (a->xdi_mbox.data) {
          int i;
          dword* data = (dword*)a->xdi_mbox.data;

          for (i = 0; i < 8; i++) {
            *data++ = a->resources.pci.bar[i];
          }
          *data++ = (dword)a->resources.pci.irq;
          a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
          ret = 0;
        }
      }
      break;

    case DIVA_XDI_UM_CMD_GET_CARD_STATE:
      if (!a->xdi_adapter.ControllerNumber) {
        a->xdi_mbox.data_length = sizeof(dword);
        a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length);
        if (a->xdi_mbox.data) {
          dword* data = (dword*)a->xdi_mbox.data;
          if (!a->xdi_adapter.ram || !a->xdi_adapter.reset ||
              !a->xdi_adapter.cfg) {
            *data = 3;
          } else if (a->xdi_adapter.trapped) {
            *data = 2;
          } else if (a->xdi_adapter.Initialized) {
            *data = 1;
          } else {
            *data = 0;
          }
          a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
          ret = 0;
        }
      }
      break;

    case DIVA_XDI_UM_CMD_WRITE_FPGA:
      if (!a->xdi_adapter.ControllerNumber) {
        ret = diva_analog_write_fpga_image (a, (byte*)&cmd[1], 
                                          cmd->command_data.write_fpga.image_length);
      }
      break;

    case DIVA_XDI_UM_CMD_RESET_ADAPTER:
      if (!a->xdi_adapter.ControllerNumber) {
        ret = diva_analog_reset_adapter (a);
      }
      break;

    case DIVA_XDI_UM_CMD_WRITE_SDRAM_BLOCK:
      ret = diva_analog_write_sdram_block (&a->xdi_adapter,
                                           cmd->command_data.write_sdram.offset,
                                           (byte*)&cmd[1],
                                           cmd->command_data.write_sdram.length,
                                           a->xdi_adapter.MemorySize);
      break;

    case DIVA_XDI_UM_CMD_START_ADAPTER:
      ret = diva_analog_start_adapter (&a->xdi_adapter,
                                       cmd->command_data.start.offset,
                                       cmd->command_data.start.features);
      break;

    case DIVA_XDI_UM_CMD_SET_PROTOCOL_FEATURES:
      a->xdi_adapter.features = cmd->command_data.features.features;
      a->xdi_adapter.a.protocol_capabilities = a->xdi_adapter.features;
      DBG_TRC(("Set raw protocol features (%08x)", a->xdi_adapter.features))
      ret = 0;
      break;

    case DIVA_XDI_UM_CMD_ADAPTER_TEST:
      if (a->xdi_adapter.DivaAdapterTestProc) {
        a->xdi_adapter.AdapterTestMask = cmd->command_data.test.test_command;
        if ((*(a->xdi_adapter.DivaAdapterTestProc))(&a->xdi_adapter)) {
          ret = -1;
        } else {
          ret = 0;
        }
      } else {
        ret = -1;
      }
      a->xdi_adapter.AdapterTestMask = 0;
      break;

    case DIVA_XDI_UM_CMD_ALLOC_DMA_DESCRIPTOR:
      if (a->xdi_adapter.dma_map) {
        a->xdi_mbox.data_length = sizeof(diva_xdi_um_cfg_cmd_data_alloc_dma_descriptor_t);
        if ((a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length))) {
          diva_xdi_um_cfg_cmd_data_alloc_dma_descriptor_t* p = 
            (diva_xdi_um_cfg_cmd_data_alloc_dma_descriptor_t*)a->xdi_mbox.data;
          int nr = diva_alloc_dma_map_entry ((struct _diva_dma_map_entry*)a->xdi_adapter.dma_map);
          unsigned long dma_magic;
          void* local_addr;

          if (nr >= 0) {
            diva_get_dma_map_entry ((struct _diva_dma_map_entry*)a->xdi_adapter.dma_map,
                                    nr,
                                    &local_addr, &dma_magic);
            p->nr   = (dword)nr;
            p->low  = (dword)dma_magic;
            p->high = 0;
          } else {
            p->nr   = 0xffffffff;
            p->high = 0;
            p->low  = 0;
          }
          a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
          ret = 0;
        }
      }
      break;

    case DIVA_XDI_UM_CMD_STOP_ADAPTER:
      ret = diva_analog_stop_adapter (a);
      break;

    case DIVA_XDI_UM_CMD_READ_XLOG_ENTRY:
      ret = diva_card_read_xlog (a);
      break;

    case DIVA_XDI_UM_CMD_READ_SDRAM:
      if (a->xdi_adapter.Address) {
        if ((a->xdi_mbox.data_length = cmd->command_data.read_sdram.length)) {
          if ((a->xdi_mbox.data_length+cmd->command_data.read_sdram.offset) < a->xdi_adapter.MemorySize) {
            a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length);
            if (a->xdi_mbox.data) {
              byte* src = a->xdi_adapter.Address;
              byte* dst = a->xdi_mbox.data;
              dword len = a->xdi_mbox.data_length;

              src += cmd->command_data.read_sdram.offset;

              while (len--) {
                *dst++ = *src++;
              }
              a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
              ret = 0;
            }
          }
        }
      }
      break;

		case DIVA_XDI_UM_CMD_GET_HW_INFO_STRUCT:
			a->xdi_mbox.data_length = sizeof(a->xdi_adapter.hw_info);
			if ((a->xdi_mbox.data = diva_os_malloc (0, a->xdi_mbox.data_length)) != 0) {
				memcpy (a->xdi_mbox.data, a->xdi_adapter.hw_info, sizeof(a->xdi_adapter.hw_info));
				a->xdi_mbox.status = DIVA_XDI_MBOX_BUSY;
				ret = 0;
			}
			break;

    default:
      DBG_ERR(("A: A(%d) invalid cmd=%d", a->controller, cmd->command))
  }

  return (ret);
}

void
diva_os_prepare_analog_functions (PISDN_ADAPTER IoAdapter)
{
}

static int
diva_analog_write_fpga_image (diva_os_xdi_adapter_t* a, byte* data, dword length)
{
  int ret;

  diva_xdiLoadFileFile = data;
  diva_xdiLoadFileLength = length;

  ret = analog_FPGA_download (&a->xdi_adapter);

  diva_xdiLoadFileFile = 0;
  diva_xdiLoadFileLength = 0;

  return (ret ? 0 : -1);
}

static int
diva_analog_reset_adapter (struct _diva_os_xdi_adapter* a)
{
	PISDN_ADAPTER IoAdapter = &a->xdi_adapter;
  byte Bus, Slot;
  void* hdev;
  byte irq;

  if (!IoAdapter->Address || !IoAdapter->reset) {
    return (-1);
  }
  if (IoAdapter->Initialized) {
    DBG_ERR(("A: A(%d) can't reset ANALOG adapter - please stop first", IoAdapter->ANum))
    return (-1);
  }

  /*
    Reset erases interrupt setting from the
    PCI configuration space
  */
  Bus = a->resources.pci.bus;
  Slot = a->resources.pci.func;
  hdev = a->resources.pci.hdev;

  PCIread (Bus, Slot, 0x3c, &irq, sizeof(irq), hdev) ;
  DBG_TRC(("SAVE IRQ LINE: %02x", irq))

  (*(IoAdapter->rstFnc))(IoAdapter);

  PCIwrite (Bus, Slot, 0x3c, &irq, sizeof(irq), hdev) ;
  PCIread  (Bus, Slot, 0x3c, &irq, sizeof(irq), hdev) ;
  DBG_TRC(("RESTORED IRQ LINE: %02x", irq))

  /*
    Forget all entities on all adapters
  */
  IoAdapter->e_count =  0;
  if (IoAdapter->e_tbl) {
    memset (IoAdapter->e_tbl, 0x00, IoAdapter->e_max * sizeof(E_INFO));
  }
  IoAdapter->head	= 0;
  IoAdapter->tail = 0;
  IoAdapter->assign = 0;
  IoAdapter->trapped = 0;

  memset (&IoAdapter->a.IdTable[0],              0x00, sizeof(IoAdapter->a.IdTable));
  memset (&IoAdapter->a.IdTypeTable[0],          0x00, sizeof(IoAdapter->a.IdTypeTable));
  memset (&IoAdapter->a.FlowControlIdTable[0],   0x00, sizeof(IoAdapter->a.FlowControlIdTable));
  memset (&IoAdapter->a.FlowControlSkipTable[0], 0x00, sizeof(IoAdapter->a.FlowControlSkipTable));
  memset (&IoAdapter->a.misc_flags_table[0],     0x00, sizeof(IoAdapter->a.misc_flags_table));
  memset (&IoAdapter->a.rx_stream[0],            0x00, sizeof(IoAdapter->a.rx_stream));
  memset (&IoAdapter->a.tx_stream[0],            0x00, sizeof(IoAdapter->a.tx_stream));

  if (IoAdapter->dma_map) {
    diva_reset_dma_mapping (IoAdapter->dma_map);
  }

  return (0);
}


static int
diva_analog_write_sdram_block (PISDN_ADAPTER IoAdapter,
                               dword address,
                               const byte* data,
                               dword length,
                               dword limit)
{
  byte* mem = IoAdapter->Address;

  if (((address + length) >= limit) || !mem) {
    DBG_ERR(("A(%d) write ANALOG address=0x%08lx", IoAdapter->ANum, address+length))
    return (-1);
  }
  mem += address;

  while (length--) {
    *mem++ = *data++;
  }

  return (0);
}

static int
diva_analog_start_adapter (PISDN_ADAPTER IoAdapter,
                         dword start_address,
                         dword features)
{
  volatile word *signature;
  int started = 0;
  int i;

  /*
    start adapter
  */
  start_analog_hardware (IoAdapter);

  /*
    wait for signature in shared memory (max. 3 seconds)
  */
  signature = (volatile word *)(&IoAdapter->ram[0x1E]) ;

  for ( i = 0 ; i < 300 ; ++i ) {
    diva_os_wait (10) ;
    if (signature[0] == 0x4447) {
      DBG_TRC(("Protocol startup time %d.%02d seconds", (i / 100), (i % 100) ))
      started = 1;
      break;
    }
  }

  if (!started) {
    DBG_FTL(("%s: Adapter selftest failed, signature=%04x",
              IoAdapter->Properties.Name, signature[0]))
      (*(IoAdapter->trapFnc))(IoAdapter);
    IoAdapter->stop(IoAdapter);
    return (-1);
  }

  IoAdapter->Initialized = 1;
  IoAdapter->IrqCount = 0;

  if (check_analog_interrupt (IoAdapter)) {
    DBG_ERR(("A(%d) interrupt test failed", IoAdapter->ANum))
    IoAdapter->Initialized = 0;
    IoAdapter->stop(IoAdapter);
    return (-1);
  }

  IoAdapter->Properties.Features = (word)features;
  diva_xdi_display_adapter_features (IoAdapter->ANum);

  DBG_LOG(("A(%d) ANALOG adapter successfull started", IoAdapter->ANum))

  diva_xdi_didd_register_adapter (IoAdapter->ANum);

  return (0);
}

static int
check_analog_interrupt (PISDN_ADAPTER IoAdapter)
{
#ifdef	SUPPORT_INTERRUPT_TEST_ON_4BRI
  int i ;
  ADAPTER *a = &IoAdapter->a ;

  IoAdapter->IrqCount = 0 ;

  IoAdapter->reset[PLX9054_INTCSR] = PLX9054_INT_ENABLE ;
  /*
    interrupt test
  */
  a->ReadyInt = 1 ;
  a->ram_out (a, &PR_RAM->ReadyInt, 1) ;

  for ( i = 100 ; !IoAdapter->IrqCount && (i-- > 0) ; diva_os_wait(10)) ;

  return ((IoAdapter->IrqCount > 0) ? 0 : -1);
#else
  dword volatile *Irq ;
  /*
    Reset on-board interrupt register
  */
  IoAdapter->IrqCount = 0 ;
  Irq = (dword volatile *)(&IoAdapter->ctlReg[MQ2_BREG_IRQ_TEST]);

  *Irq = MQ_IRQ_REQ_OFF ;

  IoAdapter->reset[PLX9054_INTCSR] = PLX9054_INT_ENABLE;

  diva_os_wait(100);

  return (0) ;
#endif	/* SUPPORT_INTERRUPT_TEST_ON_4BRI */
}

static void
diva_analog_clear_interrupts (diva_os_xdi_adapter_t* a)
{
  PISDN_ADAPTER IoAdapter = &a->xdi_adapter;

  /*
    clear any pending interrupt
  */
  IoAdapter->disIrq (IoAdapter) ;

  IoAdapter->tst_irq (&IoAdapter->a) ;
  IoAdapter->clr_irq (&IoAdapter->a) ;
  IoAdapter->tst_irq (&IoAdapter->a) ;
}

static int
diva_analog_stop_adapter (diva_os_xdi_adapter_t* a)
{
  PISDN_ADAPTER IoAdapter = &a->xdi_adapter;
  int i;

  if (!IoAdapter->ram) {
    return (-1);
  }

  if (!IoAdapter->Initialized) {
    DBG_ERR(("A(%d) can't stop ANALOG adapter - not running", IoAdapter->ANum))
    return (-1); /* nothing to stop */
  }

  IoAdapter->Initialized = 0;

  /*
    Disconnect Adapters from DIDD
  */
  diva_xdi_didd_remove_adapter(IoAdapter->ANum);

  i = 100;

  /*
    Stop interrupts
  */
  a->clear_interrupts_proc = diva_analog_clear_interrupts;
  IoAdapter->a.ReadyInt = 1;
  IoAdapter->a.ram_inc (&IoAdapter->a, &PR_RAM->ReadyInt) ;
  do {
    diva_os_sleep (10);
  } while (i-- && a->clear_interrupts_proc);
  if (a->clear_interrupts_proc) {
    diva_analog_clear_interrupts (a);
    a->clear_interrupts_proc = 0;
    DBG_ERR(("A(%d) no final interrupt from ANALOG adapter", IoAdapter->ANum))
  }
  IoAdapter->a.ReadyInt = 0;
  /*
    kill pending dpcs
  */
  diva_os_cancel_soft_isr (&IoAdapter->req_soft_isr);
  diva_os_cancel_soft_isr (&IoAdapter->isr_soft_isr);

  /*
    Stop and reset adapter
  */
  IoAdapter->stop (IoAdapter) ;

  return (0);
}

